@@ -1,16 +1,18 @@
-#   Copyright 2021 AUI, Inc. Washington DC, USA
+#  CASA Next Generation Infrastructure
+#  Copyright (C) 2021 AUI, Inc. Washington DC, USA
 #
-#   Licensed under the Apache License, Version 2.0 (the "License");
-#   you may not use this file except in compliance with the License.
-#   You may obtain a copy of the License at
+#  This program is free software: you can redistribute it and/or modify
+#  it under the terms of the GNU General Public License as published by
+#  the Free Software Foundation, either version 3 of the License, or
+#  (at your option) any later version.
 #
-#       http://www.apache.org/licenses/LICENSE-2.0
+#  This program is distributed in the hope that it will be useful,
+#  but WITHOUT ANY WARRANTY; without even the implied warranty of
+#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+#  GNU General Public License for more details.
 #
-#   Unless required by applicable law or agreed to in writing, software
-#   distributed under the License is distributed on an "AS IS" BASIS,
-#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
-#   See the License for the specific language governing permissions and
-#   limitations under the License.
+#  You should have received a copy of the GNU General Public License
+#  along with this program.  If not, see <https://www.gnu.org/licenses/>.
 
 #################################
 # Helper File
@@ -22,14 +24,16 @@
 import xarray as xr
 import typing
 
-def _apply_coord_remap(xds : xr.Dataset, coord_name : str, map_func) -> xr.Dataset:
+def apply_coord_remap(xds : xr.Dataset, coord_name : str, map_func) -> xr.Dataset:
     """Apply the map_func to the values in the given coordinate"""
+    import numpy as np
+
     # get the new values
     old_vals = xds[coord_name].values
-    new_vals = [map_func(x) for x in old_vals]
+    new_vals = np.array([map_func(x) for x in old_vals], old_vals.dtype)
 
     # assign the coordinate
-    if (coord_name in xds.dims):
+    if coord_name in xds.dims:
         # assign as dimensional coordinate
         ret = xds.drop_vars([ coord_name ])
         ret = ret.assign_coords({ coord_name: new_vals })
@@ -41,20 +45,21 @@ def _apply_coord_remap(xds : xr.Dataset, coord_name : str, map_func) -> xr.Datas
         ret = ret.set_coords([ coord_name ]) # promote the new data_var "coord_name" to be a coordinate
     return ret
 
-def _apply_data_var_remap(xds : xr.Dataset, var_name : str, map_func) -> xr.Dataset:
+def apply_data_var_remap(xds : xr.Dataset, var_name : str, map_func) -> xr.Dataset:
     """Apply the map_func to the values in the given data_var"""
     import numpy as np
     def mb(array):
         vals = array.values
-        if (len(vals) > 0):
-            vals = np.vectorize(map_func)(vals)
-        return xr.DataArray(data=vals, coords=array.coords, dims=array.dims, name=array.name, attrs=array.attrs)
+        newvals = np.ndarray(vals.shape, vals.dtype)
+        if len(vals) > 0:
+            newvals = np.vectorize(map_func, [vals.dtype])(vals)
+        return xr.DataArray(data=newvals, coords=array.coords, dims=array.dims, name=array.name, attrs=array.attrs)
 
     assert (isinstance(xds[var_name], xr.DataArray)), f"######### ERROR: trying to remap the data variable {var_name} which is a {type(xds[var_name])} but a {xr.DataArray} was expected!"
     var_val = xds[var_name].map_blocks(mb)
     return xds.assign({var_name: var_val})
 
-def _get_subtable_dimcoords_or_primcoords(sub : xr.Dataset, subtable_name : str) -> typing.List[str]:
+def get_subtable_dimcoords_or_primcoords(sub : xr.Dataset, subtable_name : str) -> typing.List[str]:
     """Get the dimension coordinate and primary key coordinate names from the given subtable"""
     import numpy as np
     from cngi._utils._mxds_ops import get_subtable_primary_key_names
@@ -93,6 +98,7 @@ def _get_subtable_matching_dimcoords(sub0: xr.Dataset, sub1 : xr.Dataset, subtab
         Keys are the name of the coordinates. Values are dictionaries that map
         from the sub1 coordinate values to the sub0 coordinate values.
     """
+    import warnings
     # check parameters
     assert(isinstance(sub0, xr.Dataset)), f"######### ERROR: subtable {subtable_name} must be a dataset!"
     assert(isinstance(sub1, xr.Dataset)), f"######### ERROR: subtable {subtable_name} must be a dataset!"
@@ -102,15 +108,15 @@ def _get_subtable_matching_dimcoords(sub0: xr.Dataset, sub1 : xr.Dataset, subtab
         return {}
 
     # get the dimensional/primary coordinates
-    check_coords0 = _get_subtable_dimcoords_or_primcoords(sub0, subtable_name)
-    check_coords1 = _get_subtable_dimcoords_or_primcoords(sub1, subtable_name)
+    check_coords0 = get_subtable_dimcoords_or_primcoords(sub0, subtable_name)
+    check_coords1 = get_subtable_dimcoords_or_primcoords(sub1, subtable_name)
 
     # Check for equality in dependent coordinates/data_vars.
     coords_and_vars = list(sub1.coords) + list(sub1.data_vars)
     ret = {}
     for coord_name in check_coords1:
         # don't worry about dimcoords that aren't in sub0
-        if (coord_name not in check_coords0):
+        if coord_name not in check_coords0:
             continue
 
         # build the dictionary of vals to check
@@ -134,7 +140,7 @@ def _get_subtable_matching_dimcoords(sub0: xr.Dataset, sub1 : xr.Dataset, subtab
                 continue
 
             # get dimensions of the this coordinate/variable
-            # only continue if the dimcoord is the first dimension
+            # stop if the dimcoord is not the first dimension
             vals0 = sub0[name]
             vals1 = sub1[name]
             dims0 = list(vals0.dims)
@@ -143,11 +149,19 @@ def _get_subtable_matching_dimcoords(sub0: xr.Dataset, sub1 : xr.Dataset, subtab
                 continue
             assert(dims0[0] == dimname), f"######### ERROR: subtables structure mismatch! {subtable_name}0.{name}.dims:{dims0}, {subtable_name}1.{name}.dims:{dims1}!"
 
-            # check for equality (should be delayed)
+            # check for equality
             for coord_val in sub1[coord_name].values:
                 if not vals_match[coord_val]:
                     continue # no point in checking
-                vals_match[coord_val] &= vals1[coord_val].broadcast_equals(vals0[coord_val])
+                # Do any necessary computation here to hopefully catch errors
+                # without getting into any messy broadcast_equals code.
+                vals0[coord_val].compute()
+                vals1[coord_val].compute()
+                # not a lot that we can do about the way dask does comparisons
+                # https://stackoverflow.com/questions/40659212/futurewarning-elementwise-comparison-failed-returning-scalar-but-in-the-futur
+                with warnings.catch_warnings():
+                    warnings.simplefilter(action='ignore', category=FutureWarning)
+                    vals_match[coord_val] &= vals1[coord_val].broadcast_equals(vals0[coord_val])
 
         # add all matching dimcoord values to the returned match dictionary
         for coord_val in sub1[coord_name].values:
@@ -205,8 +219,8 @@ def _get_subtable_dimcoord_remap(sub0: xr.Dataset, sub1 : xr.Dataset, subtable_n
     assert(isinstance(sub1, xr.Dataset)), f"######### ERROR: subtable {subtable_name} must be a dataset!"
 
     # get the dimensional/primary coordinates
-    check_coords0 = _get_subtable_dimcoords_or_primcoords(sub0, subtable_name)
-    check_coords1 = _get_subtable_dimcoords_or_primcoords(sub1, subtable_name)
+    check_coords0 = get_subtable_dimcoords_or_primcoords(sub0, subtable_name)
+    check_coords1 = get_subtable_dimcoords_or_primcoords(sub1, subtable_name)
     if print_warnings and (check_coords0 != check_coords1):
         print("Warning: subtable primary keys do not match. Is it possible that sub0 and sub1 are from different converter versions?")
         print(f"sub0 primary keys: {check_coords0}")
@@ -291,7 +305,7 @@ def _remap_subtable_coords_and_vals(sub1 : xr.Dataset, subtable_name : str, coor
     assert(isinstance(sub1, xr.Dataset)), f"######### ERROR: subtable {subtable_name} must be a dataset!"
 
     # get the list of non-reference coordinates
-    nonref_coord_names = _get_subtable_dimcoords_or_primcoords(sub1, subtable_name)
+    nonref_coord_names = get_subtable_dimcoords_or_primcoords(sub1, subtable_name)
 
     # assign the new coordinate values in sub1
     for coord_name in coords_vals_remap.keys():
@@ -308,7 +322,7 @@ def _remap_subtable_coords_and_vals(sub1 : xr.Dataset, subtable_name : str, coor
         map_func = lambda x: x if x not in map_vals else map_vals[x]
 
         # apply the new coordinate
-        sub1 = _apply_coord_remap(sub1, coord_name, map_func)
+        sub1 = apply_coord_remap(sub1, coord_name, map_func)
 
     # apply the relational_ids_map to sub1
     for var_name in relational_ids_map:
@@ -322,13 +336,13 @@ def _remap_subtable_coords_and_vals(sub1 : xr.Dataset, subtable_name : str, coor
 
         # apply the mapping
         if var_name in sub1.coords:
-            sub1 = _apply_coord_remap(sub1, var_name, map_func)
+            sub1 = apply_coord_remap(sub1, var_name, map_func)
         elif var_name in sub1.data_vars:
-            sub1 = _apply_data_var_remap(sub1, var_name, map_func)
+            sub1 = apply_data_var_remap(sub1, var_name, map_func)
 
     return sub1
 
-def _build_mxds_coords(mxds : xr.Dataset, subtables : dict):
+def build_mxds_coords(mxds : xr.Dataset, subtables : typing.Dict[str, xr.Dataset]):
     from cngi._utils._mxds_ops import get_subtable_primary_key_names
 
     # make a copy of the current coordinates, as a dictionary so that it can be easily updated
@@ -400,9 +414,9 @@ def append_mxds_subtables(mxds0 : xr.Dataset, mxds1 : xr.Dataset, matchtype="exa
     ----------------
     pseudocode: ret = mxds0.append(mxds1.coords).append(mxds1.attrs.sel(!"xds*"))
 
-    This function makes heavy use of append_xds_subtable to append subtables
-    from mxds1 to mxds0 in a sort of round-robin style. The resulting
-    subtables are merged into a new mxds to be returned.
+    This function makes heavy use of _remap_subtable_coords_and_vals to update
+    references from mxds1 to mxds0 in a sort of round-robin style. The
+    resulting subtables are merged into a new mxds to be returned.
 
     The new mxds maintains all dims and coords of mxds0, with possibly expanded
     values for the dimensional coordinates.
@@ -465,8 +479,6 @@ def append_mxds_subtables(mxds0 : xr.Dataset, mxds1 : xr.Dataset, matchtype="exa
     keys_changed = True
     fresh_subs = subs.copy()
     while keys_changed:
-        keys_changed = False
-
         # Get the coordinate remapping for coordinate relationships between mxds1
         # subtables, and check that there are no dimensional coordinates conflicts
         # between any of the subtables in mxds1.
@@ -510,7 +522,7 @@ def append_mxds_subtables(mxds0 : xr.Dataset, mxds1 : xr.Dataset, matchtype="exa
             attrs[sn] = subs[sn]
 
     # update the global coordinates to reflect the new coordinate values in the subtables
-    coords = _build_mxds_coords(mxds0, attrs)
+    coords = build_mxds_coords(mxds0, attrs)
 
     # merge mxds0 and mxds1
     ret = xr.Dataset(data_vars=mxds0.data_vars, coords=coords, attrs=attrs)
@@ -518,145 +530,23 @@ def append_mxds_subtables(mxds0 : xr.Dataset, mxds1 : xr.Dataset, matchtype="exa
 
     return ret, ids_map
 
-def extract_xds_with_subtables(mxds : xr.Dataset, xds_names : typing.Union[str, typing.List[str]]) -> xr.Dataset:
-    """Pull the xds visibilites out with the mxds, preserving only that information in the subtables that is related to the given xds.
+def gen_keyname_variants(keyname : str) -> typing.List[str]:
+    """The dimensional coordinate names and primary coordinate names from the
+    mxds subtables are mangled in their usage in the main xds visibility tables.
+    This function provides all possible mangled versions that can be matched
+    against for doing comparisons.
 
-    Parameters
-    ----------
-    mxds: xarray.Dataset
-        The multi-xds dataset to pull data out of.
-    xds_name: str or list
-        Name(s) of the visibilities dataset. Should be of the form "xds*"
-
-    Returns
-    -------
-    xarray.Dataset
-        A new mxds, which includes just the xds_name visibility Dataset and the
-        related information from the subtables.
+    Example keynames: antenna_id, beam_id, feed_id, spectral_window_id, ns_ws_station_id
+    Example variants: "pol_id", "spw_id", "ANTENNA1", "ANTENNA2", "ARRAY_ID", "FEED1"
     """
-    import numpy as np
-    from cngi._utils._mxds_ops import get_subtables, get_subtable_primary_key_names, check_mxds_subtable_ref_ids, assign_dimensions_for_primary_coordinates
+    knvariants = []
 
-    # get a list of the main tables
-    main_tables = [] # type: typing.List[xr.Dataset]
-    if isinstance(xds_names, str):
-        xds_names = [xds_names]
-    for xds_name in xds_names:
-        assert ("xds" in xds_name), f"######### ERROR: xds_name must reference a main table! Name should contain \"xds\" but is instead {xds_name}!"
-        assert (xds_name in mxds.attrs), f"######### ERROR: main table {xds_name} does not appear in the mxds list of attrs!"
-        main = mxds.attrs[xds_name] # type: xr.Dataset
-        assert (isinstance(main, xr.Dataset)), f"######### ERROR: xds visibilities table must be a Dataset but is instead a {type(main)}!"
-        main_tables.append(main)
+    knl = keyname.lower()
+    for kn1 in [knl, knl.replace("spectral_window", "spw")]:
+        for kn2 in [kn1, kn1.replace("_id", "")]:
+            for kn3 in [kn2, kn2.upper(), kn2+'s', kn2.upper()+'S']:
+                knvariants.append(kn3)
+            for kn3 in [kn2+'1', kn2+'2', kn2.upper()+'1', kn2.upper()+'2']:
+                knvariants.append(kn3)
 
-    # make a copy of the subtables, as a dictionary so that it can be easily updated
-    # exclude all xds visibilities other than the desired main tables
-    attrs = {} # type: typing.Dict[str, xr.Dataset]
-    for sn in mxds.attrs: # sn = subtable_name
-        if ("xds" not in sn) or (sn in xds_names):
-            attrs[sn] = mxds.attrs[sn]
-
-    # get the list of subtables, and the list of key coordinates used for indexing those subtables
-    # example subtables: ANTENNA, ASDM_ANTENNA, FEED, WEATHER
-    # example key coordinates: antenna_id, beam_id, feed_id, spectral_window_id -> spw_id, ns_ws_station_id
-    subnames = get_subtables(mxds)
-    sub_keynames = {}
-    keynames = [] # type: typing.List[str]
-    for sn in subnames:
-        sub_kns = _get_subtable_dimcoords_or_primcoords(attrs[sn], sn)
-        keynames += sub_kns
-        sub_keynames[sn] = sub_kns
-    keynames = list(np.unique(keynames))
-
-    # get the list of key values to keep, based off of the main tables
-    used_keyvals = {}
-    used_knvariants = []
-    for kn in keynames:
-        # get a list of variant keynames that could be used in the main tables
-        # example variants: "pol_id", "spw_id", "ANTENNA1", "ANTENNA2", "ARRAY_ID", "FEED1"
-        knvariants = []
-        knl = kn.lower()
-        for kn1 in [kn, knl.replace("spectral_window", "spw")]:
-            for kn2 in [kn1, kn1.replace("_id", "")]:
-                for kn3 in [kn2, kn2.upper(), kn2+'s', kn2.upper()+'S']:
-                    knvariants.append(kn3)
-                for kn3 in [kn2+'1', kn2+'2', kn2.upper()+'1', kn2.upper()+'2']:
-                    knvariants.append(kn3)
-
-        # find the used values
-        used = []
-        for main in main_tables:
-            for knvariant in knvariants:
-                if (knvariant not in main.coords) and (knvariant not in main.data_vars):
-                    continue
-                used_knvariants.append(knvariant)
-                vals = np.unique(main[knvariant].values) # get unique values along each dimension
-                tmpused = np.unique(vals.flatten()) # flatten to a single dimension and get unique values along that single dimension
-                tmpused = list(filter(lambda x: not np.isnan(x), tmpused))
-                used += tmpused
-        if len(used) == 0:
-            used_keyvals[kn] = None
-        else:
-            used_keyvals[kn] = np.unique(used)
-    # print(keynames)
-    # print(np.unique(used_knvariants))
-
-    # build a new set of subtables with trimmed values
-    alldrops = {}
-    for sn in subnames:
-        sub = attrs[sn]
-
-        # we don't know how to trim down tables that don't have keys
-        if len(sub_keynames[sn]) == 0:
-            # automatically keep this entire table
-            continue
-
-        # find the used dimensions of the subtable based on its keys
-        used_dims = {}
-        for dn in sub.dims:
-            # limit this dimension based on which values are used by the key coordinates
-            # TODO how to find used dimension values of multi-dimension keys?
-            for kn in sub_keynames[sn]:
-                coord = sub[kn]
-                if used_keyvals[kn] is None:
-                    continue
-                if dn not in coord.dims:
-                    continue
-                if dn not in used_dims:
-                    used_dims[dn] = []
-                # find the matching dimension values to the those used key coordinate values
-                used = []
-                for dimval in sub[dn].values:
-                    keyval = coord.sel({dn:[dimval]}).values[0]
-                    if keyval in used_keyvals[kn]:
-                        used.append(dimval)
-                used_dims[dn] += used
-
-            # if this dimension does not appear in any keynames, then assume the entire dimension is used
-            if dn not in used_dims:
-                used_dims[dn] = sub[dn].values
-
-        # find dimension values that aren't used, to be dropped
-        dropvals = {}
-        for dn in sub.dims:
-            used = used_dims[dn]
-            unused = list(filter(lambda v: v not in used, sub[dn].values))
-            if len(unused) > 0:
-                dropvals[dn] = unused
-        # TODO remove
-        alldrops[sn] = dropvals
-
-        # drop unused dimension values
-        if len(dropvals) > 0:
-            attrs[sn] = sub.drop_sel(dropvals)
-
-    # update the global coordinates to reflect the new coordinate values in the subtables
-    coords = _build_mxds_coords(mxds, attrs)
-
-    # for kn in keynames:
-    #     if used_keyvals[kn] is not None:
-    #         print(f"{kn} used values: {np.sort(used_keyvals[kn])}")
-    # for sn in alldrops:
-    #     print(f"{sn} dropped values: {alldrops[sn]}")
-
-    # create the new mxds
-    return xr.Dataset(coords=coords, data_vars=mxds.data_vars, attrs=attrs)
\ No newline at end of file
+    return knvariants
